/*
 * z2.c  --  SoC audio for Aeronix Zipit Z2
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Copyright 2005 Openedhand Ltd.
 *
 * Authors: Ken McGuire <kenm@desertweyr.com>
 *	Based on spitz.c by:
 *	    Liam Girdwood <liam.girdwood@wolfsonmicro.com>
 *          Richard Purdie <richard@openedhand.com>
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  Revision history
 *    21st Dec 2008   Zipit Z2 version.    
 *    30th Nov 2005   Initial version.
 *
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/mach-types.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/hardware.h>
#include <asm/mach-types.h>
#include "../codecs/wm8750.h"
#include "pxa2xx-pcm.h"
#include "pxa2xx-i2s.h"

#define Z2_HP        0
#define Z2_MIC       1
#define Z2_LINE      2
#define Z2_HEADSET   3
#define Z2_HP_OFF    4
#define Z2_SPK_ON    0
#define Z2_SPK_OFF   1

 /* audio clock in Hz - rounded from 12.235MHz */
#define Z2_AUDIO_CLOCK 12288000

static int z2_jack_func;
static int z2_spk_func;

static void z2_ext_control(struct snd_soc_codec *codec)
{
	if (z2_spk_func == Z2_SPK_ON)
		snd_soc_dapm_set_endpoint(codec, "Ext Spk", 1);
	else
		snd_soc_dapm_set_endpoint(codec, "Ext Spk", 0);

	/* set up jack connection */
	switch (z2_jack_func) {
	case Z2_HP:
		/* enable and unmute hp jack, disable mic bias */
		snd_soc_dapm_set_endpoint(codec, "Headset Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Mic Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Line Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Headphone Jack", 1);
		break;
	case Z2_MIC:
		/* enable mic jack and bias, mute hp */
		snd_soc_dapm_set_endpoint(codec, "Headphone Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Headset Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Line Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Mic Jack", 1);
		break;
	case Z2_LINE:
		/* enable line jack, disable mic bias and mute hp */
		snd_soc_dapm_set_endpoint(codec, "Headphone Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Headset Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Mic Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Line Jack", 1);
		break;
	case Z2_HEADSET:
		/* enable and unmute headset jack enable mic bias, mute L hp */
		snd_soc_dapm_set_endpoint(codec, "Headphone Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Mic Jack", 1);
		snd_soc_dapm_set_endpoint(codec, "Line Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Headset Jack", 1);
		break;
	case Z2_HP_OFF:

		/* jack removed, everything off */
		snd_soc_dapm_set_endpoint(codec, "Headphone Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Headset Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Mic Jack", 0);
		snd_soc_dapm_set_endpoint(codec, "Line Jack", 0);
		break;
	}
	snd_soc_dapm_sync_endpoints(codec);
}

static int z2_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_codec *codec = rtd->socdev->codec;

	/* check the jack status at stream startup */
	z2_ext_control(codec);
	return 0;
}

static int z2_hw_params(struct snd_pcm_substream *substream,
	struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_codec_dai *codec_dai = rtd->dai->codec_dai;
	struct snd_soc_cpu_dai *cpu_dai = rtd->dai->cpu_dai;
	unsigned int clk = 0;
	int ret = 0;

	switch (params_rate(params)) {
	case 8000:
	case 16000:
	case 48000:
	case 96000:
		clk = 12288000;
		break;
	case 11025:
	case 22050:
	case 44100:
		clk = 11289600;
		break;
	}

	/* set codec DAI configuration */
	ret = codec_dai->dai_ops.set_fmt(codec_dai, SND_SOC_DAIFMT_I2S |
		SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS);
	if (ret < 0)
		return ret;

	/* set cpu DAI configuration */
	ret = cpu_dai->dai_ops.set_fmt(cpu_dai, SND_SOC_DAIFMT_I2S |
		SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS);
	if (ret < 0)
		return ret;

	/* set the codec system clock for DAC and ADC */
	ret = codec_dai->dai_ops.set_sysclk(codec_dai, WM8750_SYSCLK, clk,
		SND_SOC_CLOCK_IN);
	if (ret < 0)
		return ret;

	/* set the I2S system clock as input (unused) */
	ret = cpu_dai->dai_ops.set_sysclk(cpu_dai, PXA2XX_I2S_SYSCLK, 0,
		SND_SOC_CLOCK_IN);
	if (ret < 0)
		return ret;

	return 0;
}

static struct snd_soc_ops z2_ops = {
	.startup = z2_startup,
	.hw_params = z2_hw_params,
};

static int z2_get_jack(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = z2_jack_func;
	return 0;
}

static int z2_set_jack(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);

	if (z2_jack_func == ucontrol->value.integer.value[0])
		return 0;

	z2_jack_func = ucontrol->value.integer.value[0];
	z2_ext_control(codec);
	return 1;
}

static int z2_get_spk(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = z2_spk_func;
	return 0;
}

static int z2_set_spk(struct snd_kcontrol *kcontrol,
	struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_codec *codec =  snd_kcontrol_chip(kcontrol);

	if (z2_spk_func == ucontrol->value.integer.value[0])
		return 0;

	z2_spk_func = ucontrol->value.integer.value[0];
	z2_ext_control(codec);
	return 1;
}
/*
static int z2_mic_bias(struct snd_soc_dapm_widget *w, int event)
{
	if (machine_is_borzoi() || machine_is_z2()) {
		if (SND_SOC_DAPM_EVENT_ON(event))
			set_scoop_gpio(&z2scoop2_device.dev,
				Z2_SCP2_MIC_BIAS);
		else
			reset_scoop_gpio(&z2scoop2_device.dev,
				Z2_SCP2_MIC_BIAS);
	}

	if (machine_is_akita()) {
		if (SND_SOC_DAPM_EVENT_ON(event))
			akita_set_ioexp(&akitaioexp_device.dev,
				AKITA_IOEXP_MIC_BIAS);
		else
			akita_reset_ioexp(&akitaioexp_device.dev,
				AKITA_IOEXP_MIC_BIAS);
	}
	return 0;
}
*/
/* z2 machine dapm widgets */
static const struct snd_soc_dapm_widget wm8750_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headphone Jack", NULL),
	SND_SOC_DAPM_MIC("Mic Jack", NULL),
	SND_SOC_DAPM_SPK("Ext Spk", NULL),
	SND_SOC_DAPM_LINE("Line Jack", NULL),

	/* headset is a mic and mono headphone */
	SND_SOC_DAPM_HP("Headset Jack", NULL),
};

/* Z2 machine audio_map */
static const char *audio_map[][3] = {

	/* headphone connected to LOUT1, ROUT1 */
	{"Headphone Jack", NULL, "LOUT1"},
	{"Headphone Jack", NULL, "ROUT1"},

	/* headset connected to ROUT1 and RINPUT2 with bias (def below) */
	{"Headset Jack", NULL, "ROUT1"},

	/* ext speaker connected to LOUT2, ROUT2  */
	{"Ext Spk", NULL , "ROUT2"},
	{"Ext Spk", NULL , "LOUT2"},

	/* mic is connected to R input 2 - with bias */
	{"RINPUT2", NULL, "Mic Bias"},
	{"Mic Bias", NULL, "Mic Jack"},

	/* line is connected to R input 1 - no bias */
	{"RINPUT1", NULL, "Line Jack"},

	{NULL, NULL, NULL},
};

static const char *jack_function[] = {"Headphone", "Mic", "Line", "Headset",
	"Off"};
static const char *spk_function[] = {"On", "Off"};
static const struct soc_enum z2_enum[] = {
	SOC_ENUM_SINGLE_EXT(5, jack_function),
	SOC_ENUM_SINGLE_EXT(2, spk_function),
};

static const struct snd_kcontrol_new wm8750_z2_controls[] = {
	SOC_ENUM_EXT("Jack Function", z2_enum[0], z2_get_jack,
		z2_set_jack),
	SOC_ENUM_EXT("Speaker Function", z2_enum[1], z2_get_spk,
		z2_set_spk),
};

/*
 * Logic for a wm8750 as connected on a Z2 Device
 */
static int z2_wm8750_init(struct snd_soc_codec *codec)
{
	int i, err;

	/* NC codec pins */
//	snd_soc_dapm_set_endpoint(codec, "RINPUT1", 0);
//	snd_soc_dapm_set_endpoint(codec, "LINPUT2", 0);
//	snd_soc_dapm_set_endpoint(codec, "RINPUT2", 0);
//	snd_soc_dapm_set_endpoint(codec, "LINPUT3", 0);
//	snd_soc_dapm_set_endpoint(codec, "RINPUT3", 0);
//	snd_soc_dapm_set_endpoint(codec, "OUT3", 0);
//	snd_soc_dapm_set_endpoint(codec, "MONO", 0);

	/* Add z2 specific controls */
	for (i = 0; i < ARRAY_SIZE(wm8750_z2_controls); i++) {
		err = snd_ctl_add(codec->card,
			snd_soc_cnew(&wm8750_z2_controls[i], codec, NULL));
		if (err < 0)
			return err;
	}

	/* Add z2 specific widgets */
	for (i = 0; i < ARRAY_SIZE(wm8750_dapm_widgets); i++) {
		snd_soc_dapm_new_control(codec, &wm8750_dapm_widgets[i]);
	}

	/* Set up z2 specific audio path audio_map */
	for (i = 0; audio_map[i][0] != NULL; i++) {
		snd_soc_dapm_connect_input(codec, audio_map[i][0],
			audio_map[i][1], audio_map[i][2]);
	}

	snd_soc_dapm_sync_endpoints(codec);
	return 0;
}

/* z2 digital audio interface glue - connects codec <--> CPU */
static struct snd_soc_dai_link z2_dai = {
	.name = "wm8750",
	.stream_name = "WM8750",
	.cpu_dai = &pxa_i2s_dai,
	.codec_dai = &wm8750_dai,
	.init = z2_wm8750_init,
	.ops = &z2_ops,
};

/* z2 audio machine driver */
static struct snd_soc_machine snd_soc_machine_z2 = {
	.name = "Z2",
	.dai_link = &z2_dai,
	.num_links = 1,
};

/* z2 audio private data */
static struct wm8750_setup_data z2_wm8750_setup = {
	.i2c_address = 0x1b,
};

/* z2 audio subsystem */
static struct snd_soc_device z2_snd_devdata = {
	.machine = &snd_soc_machine_z2,
	.platform = &pxa2xx_soc_platform,
	.codec_dev = &soc_codec_dev_wm8750,
	.codec_data = &z2_wm8750_setup,
};

static struct platform_device *z2_snd_device;

static int __init z2_init(void)
{
	int ret;

//	if (!(machine_is_z2() || machine_is_borzoi() || machine_is_akita()))
//		return -ENODEV;

	z2_snd_device = platform_device_alloc("soc-audio", -1);
	if (!z2_snd_device)
		return -ENOMEM;

	platform_set_drvdata(z2_snd_device, &z2_snd_devdata);
	z2_snd_devdata.dev = &z2_snd_device->dev;
	ret = platform_device_add(z2_snd_device);

	if (ret)
		platform_device_put(z2_snd_device);

	return ret;
}

static void __exit z2_exit(void)
{
	platform_device_unregister(z2_snd_device);
}

module_init(z2_init);
module_exit(z2_exit);

MODULE_AUTHOR("Ken McGuire");
MODULE_DESCRIPTION("ALSA SoC Z2");
MODULE_LICENSE("GPL");
